(* Copyright (c) 2003 Michal Moskal <malekith@pld-linux.org>
   Licensed under the Academic Free License version 2.0 *)
let xf = Printf.sprintf

(* given "foo{ba{}r}}" return "foo{ba{}r}" *)
let get_balanced curly idx str =
  let len = String.length str in
  let rec loop lev idx =
    if idx >= len then len
    else
      match str.[idx] with
      | '{' when curly -> loop (lev + 1) (idx + 1)
      | '}' when curly && lev = 0 -> idx
      | '}' when curly -> loop (lev - 1) (idx + 1)
      | '(' when not curly -> loop (lev + 1) (idx + 1)
      | ')' when not curly && lev = 0 -> idx
      | ')' when not curly -> loop (lev - 1) (idx + 1)
      | _ -> loop lev (idx + 1)
  in String.sub str idx (loop 0 idx - idx)

let rec expand loop_lev line =
  if loop_lev > 100 || String.length line > 10000 then line
  else
    if String.contains line '%' then
      let buf = Buffer.create (String.length line + 100) in
      let len = String.length line in
      let rec loop idx =
        if idx >= len then Buffer.contents buf
        else
          match line.[idx] with
          | '%' when idx < len - 1 ->
            begin
              match line.[idx + 1] with
              (* Hack for http:// ... %20 ... in src *)
              | '0' .. '9' as c when Macro.is_global_ctx () ->
                Buffer.add_char buf '%';
                Buffer.add_char buf c;
                loop (idx + 2)
              | '%' ->
                Buffer.add_char buf '%';
                loop (idx + 2)
              | '{' ->
                let macro = get_balanced true (idx + 2) line in
                let next_ch = idx + 3 + String.length macro in
                Buffer.add_string buf (expand_macro loop_lev macro);
                loop next_ch
              | '(' ->
                let macro = get_balanced false (idx + 2) line in
                let next_ch = idx + 3 + String.length macro in
                Buffer.add_string buf (expand_shell loop_lev macro);
                loop next_ch
              | _ -> 
                let macro = My_String.chop_prefix (idx + 1) line in
                let (macro_name, _) = My_String.first_word macro in
                let just_one_word () =
                  let rec loop id =
                    if id >= len then (macro, len)
                    else
                      match line.[id] with
                      | 'a' .. 'z' 
                      | 'A' .. 'Z' 
                      | '0' .. '9' 
                      | '_' ->
                        loop (id + 1)
                      | _ ->
                        (String.sub line (idx + 1) (id - idx - 1), id)
                  in loop (idx + 1)
                in
                let (macro, len) =
                  if Macro.defined macro_name then
                    match Macro.get macro_name with
                    | Macro.Simple_macro _ 
                    | Macro.Simple_callback _ -> just_one_word ()

                    | Macro.Param_macro _
                    | Macro.Param_callback _ -> (macro, len)
                  else
                   just_one_word ()
                in
                let expanded = expand_macro loop_lev macro in
                Buffer.add_string buf expanded;
                loop len
            end
          | c -> 
            Buffer.add_char buf c; 
            loop (idx + 1)
      in loop 0
    else line
      
and expand_macro loop_lev s =
  (* Printf.eprintf "p '%s'\n" s; *)
  let gr n = Str.matched_group n s in
  let expand = expand (loop_lev + 1) in
  My_String.re_match s [
    ("^\\(!\\?\\|\\?!\\)\\([^:]+\\):\\(.*\\)$", fun () ->
       if not (Macro.defined (gr 2)) then expand (gr 3) else "");
    ("^\\?\\([^:]+\\):\\(.*\\)$", fun () ->
       if Macro.defined (gr 1) then expand (gr 2) else "");
    ("^!\\?\\([^:]+\\)$", fun () -> "");
    ("^\\?\\([^:]+\\)$", fun () ->
      if Macro.defined (gr 1) then expand ("%{" ^ (gr 1) ^ "}") else "");
    ("^expand:\\(.*\\)$", fun () -> expand (expand (gr 1)));
    (".*", fun () ->
       let (macro, args) = My_String.first_word (My_String.strip_left s) in
       if Macro.defined macro then
         match Macro.get macro with
         | Macro.Simple_macro x -> expand x
         | Macro.Param_macro x -> 
           Macro.push_ctx (My_String.split args);
           let s = expand x in
           Macro.pop_ctx ();
           s
         | Macro.Simple_callback f -> f ()
         | Macro.Param_callback f -> f args
       else
         "%{" ^ s ^ "}")
  ]

and expand_shell loop_lev s =
  let s = expand (loop_lev + 1) s in
  if !Flags.shell then
    Shell.backticks s
  else
    (Generic_parser.err (xf "shell `%s' not allowed" s); "")

